## ----------------------------------------------------------------------------------- ##
## File:    DDA_code.R 
## Purpose: Source code to perform DDA using a simulated data example 
## Author:  WW 
## Date:    Sept 21, 2018 
## ------------------------------------------------------------------------------------## 

# -------------------------------------------------
# --- functions and libraries used in the analysis 
# -------------------------------------------------

library(energy)  # used for dcor-test
library(moments) # used for skewness/kurtosis tests
library(boot)    # used for bca bootstrap CI of moment differences
source("C:\\insert_path_here\\hsic_test.R")  # used for HSIC-test 


# compute (absolute) differences in higher moments 
abs.moment.diff <- function(dat, i)
{
  bootsamp <- dat[i,] 
  skewdiff <- abs(skewness(bootsamp[,1])) - abs(skewness(bootsamp[,2]))
  kurtdiff <- abs(kurtosis(bootsamp[,1])) - abs(kurtosis(bootsamp[,2]))
  out <- c(skewdiff, kurtdiff) 
  names(out) <- c("skewdiff", "kurtdiff")
  return(out)
} 

# ------------------------------------------------
# --- generate simulated data example 
# ------------------------------------------------

set.seed(5678)  
n <- 500
byx <- 0.5
byz <- 0.2 

x <- (rchisq(n, df = 3) - 3) / sqrt(6) 
z <- rnorm(n)
y <- byx*x + byz*z + rnorm(n)
mydata <- data.frame(x, z, y) # y = outcome, x = predictor, z = covariate
 
# residualize x and y
r_x <- as.vector(scale(resid(lm(x ~ z, data = mydata)))) 
r_y <- as.vector(scale(resid(lm(y ~ z, data = mydata)))) 


# ------------------------------------------------ 
# ---- DDA: variable distributions 
# ------------------------------------------------

agostino.test(r_x)  # skew = 2.4751, z = 13.9720, p-value < 2.2e-16
agostino.test(r_y)  # skew = 0.1010, z = 0.9331, p-value = 0.3508
anscombe.test(r_x)  # kurt = 15.037, z = 10.493, p-value < 2.2e-16
anscombe.test(r_y)  # kurt = 3.5464, z = 2.1912, p-value = 0.02844

result <- boot(dat = data.frame(r_x, r_y), statistic = abs.moment.diff, R = 2000)
# skewness difference 
boot.ci(result, index = 1, type = "bca")$t0  # 2.374069
boot.ci(result, index = 1, type = "bca")     # 95% CI ( 1.435,  3.582 )

# kurtosis difference 
boot.ci(result, index = 2, type = "bca")$t0  # 11.49071
boot.ci(result, index = 2, type = "bca")     # 95% CI ( 2.69, 23.76 )

# ------------------------------------------------
# --- DDA: residual distributions 
# ------------------------------------------------

tar <- lm(y ~ x + z, data = mydata) # target model 
alt <- lm(x ~ y + z, data = mydata) # alternative model

agostino.test(resid(tar)) # skew = -0.085628, z = -0.791610, p-value = 0.4286
agostino.test(resid(alt)) # skew = 1.8123, z = 11.7320, p-value < 2.2e-16

anscombe.test(resid(tar)) # kurt = 3.07850, z = 0.51074, p-value = 0.6095
anscombe.test(resid(alt)) # kurt = 10.2020, z = 9.0224, p-value < 2.2e-16

result <- boot(dat = data.frame(resid(alt), resid(tar)), statistic = abs.moment.diff, R = 1000)

# skewness difference  
boot.ci(result, index = 1, type = "bca")$t0  # 1.726623
boot.ci(result, index = 1, type = "bca")     # 95% CI ( 1.012,  2.952 )

# kurtosis difference
boot.ci(result, index = 2, type = "bca")$t0  # 7.123956 
boot.ci(result, index = 2, type = "bca")     # 95% CI ( 1.764, 15.514 )

# ------------------------------------------------
# --- DDA: independence properties 
# ------------------------------------------------

tar <- lm(y ~ x + z, data = mydata) # target model 
alt <- lm(x ~ y + z, data = mydata) # alternative model

# HSIC 
hsic.test(tar, mydata$x, B = 500, parallelize = TRUE, cores = 4) # HSIC = 0.13485, p-value = 0.356
hsic.test(alt, mydata$y, B = 500, parallelize = TRUE, cores = 4) # HSIC = 2.3165, p-value < 2.2e-16

# dCor 
dcor(resid(tar), mydata$x) # 0.07334285
dcov.test(resid(tar), mydata$x, R = 1000) # nV^2 = 1.0524, p-value = 0.5604
dcor(resid(alt), mydata$y) # 0.2391455
dcov.test(resid(alt), mydata$y, R = 1000) # nV^2 = 11.218, p-value = 0.000999


