## ----------------------------------------------------------------------------------------------------- ##
## File:    hsic_test.R 
## Purpose: Wrapper function to compute Sen and Sen's (2014) HSIC-test. Original source code taken 
##          from Bodhisattva Sen's webpage (see http://www.stat.columbia.edu/~bodhi/Bodhi/Publications.html)
## Author:  WW; Sen & Sen 
## Date:    Sept 26, 2016 
## ----------------------------------------------------------------------------------------------------- ##

hsic.test <- function(model, x, hx = 1, hy = 1, B = 1000, parallelize = FALSE, cores = 2){

# ----------------------------------------------------------------------------------------------
# Input:
# ----------------------------------------------------------------------------------------------
# model:       estimated linear regression model 
# x:           tested predictor of the model
# hx, hy:      bandwidth parameters in HSIC computation 
# B:           number of resamples 
# parallelize: logical value indicating whether boostrapping is performed on multiple cores
# cores:       a numeric value indicating the number of cores. Only used if parallelize = TRUE
# ----------------------------------------------------------------------------------------------

    X <- model.matrix(model)
	b <- as.matrix(coef(model))
	n <- nobs(model) 
	e <- resid(model)
	if(!is.matrix(x)) x <- as.matrix(x)

	T_hat <- HSIC(x, e, 2*hx*hx, 2*hy*hy)   ## compute test-statistic

    # Implementing the bootstrap procedure (parallelized if specified)
    
	e_0 <- e - mean(e)      ## centered residuals

    if(parallelize){

      require(doParallel)
	  cl <- makeCluster(cores)
      registerDoParallel(cl)

         T_hat_B <- foreach(icount(B), .combine = c, .export = "HSIC") %dopar% {

               idx <- sample(n, n, replace=TRUE)    # with replacement samples from the errors
        	   e_B <- e_0[idx]
    
        	   idx2 <- sample(n, n, replace=TRUE)   # with replacement samples from the predictors
        	   x_B <- x[idx2,]
        	   X_B <- X[idx2,]       

        	   yhat_B  <- X_B %*% b + e_B                              # create bootstrap response values
        	   bhat_B  <- solve(t(X_B) %*% X_B) %*% t(X_B) %*% yhat_B  # get new slope estimates 
        	   e_hat_B <- yhat_B - X_B %*% bhat_B                      # bootstrap residuals

        	   hx_B <- hx; hy_B <- hy
        	   HSIC(x_B, e_hat_B, 2*hx_B*hx_B, 2*hy_B*hy_B)
               }
    
	   stopCluster(cl)  
    
	} else {

      T_hat_B <- matrix(0, B, 1)
      for(j in 1:B){
      	    idx <- sample(n, n, replace=TRUE)                # with replacement samples from the errors
        	e_B <- e_0[idx]
    
        	idx2 <- sample(n, n, replace=TRUE)               # with replacement samples from the predictors
        	x_B <- x[idx2,]
        	X_B <- X[idx2,]       

        	yhat_B  <- X_B %*% b + e_B                               # create the bootstrap response values
        	bhat_B  <- solve(t(X_B) %*% X_B) %*% t(X_B) %*% yhat_B   # get new slope estimates 
        	e_hat_B <- yhat_B - X_B %*% bhat_B                       # bootstrap residuals

        	hx_B <- hx; hy_B <- hy
        	T_hat_B[j] <- HSIC(x_B, e_hat_B, 2*hx_B*hx_B, 2*hy_B*hy_B)

 	    }
	}
	
	pval <- mean(T_hat_B >= T_hat)
	names(T_hat) <- "HSIC"
	output <- list(statistic = T_hat, p.value = pval, 
	               method = "Hilbert-Schmidt Independence Test", data.name = deparse(formula(model)))
	class(output) <- "htest"
	return(output)
}

# compute HSIC value 

HSIC <- function(X, Y, hx, hy){
	n <- length(Y)     
    H <- diag(n) - matrix(1, n, n)/n    
    K <- exp(-as.matrix(dist(X, method = "euclidean", diag = TRUE, upper = TRUE))^2/hx)
	L <- exp(-as.matrix(dist(Y, method = "euclidean", diag = TRUE, upper = TRUE))^2/hy)
	testStat <- sum(diag(K %*% H %*% L %*% H))/n
	return(testStat)
}

